// AM_Point marker in (ZDoom) orthogonal point
class AM_Point : Actor
{
	Default
	{
		+FLATSPRITE
		Scale 0.0;
	}
	
	States
	{
	Spawn:
		APLS A -1;
		Stop;
	}
}

// ambushers event handler

class ambushers_EventHandler : EventHandler
{
	PlayerInfo player;

	int MOFLAG;
	int AMSTATUS;
	int debugTID;
	int counter;

	// find AM_Point
	AM_Point findambush(Actor mo, Actor pl)
	{
		ThinkerIterator AM_PointFinder;
		AM_PointFinder = ThinkerIterator.Create("AM_Point");
		AM_Point idpoint;
		AM_Point ptr = null;

		int am_radius = CVar.FindCvar("am_radius").GetInt();
		
		while (idpoint = AM_Point(AM_PointFinder.Next()))
		{
			double mo2id = mo.Distance2D(idpoint);

			if (mo2id < am_radius && !idpoint.CheckSight(pl) && idpoint.CheckSight(mo)) 
			{
				ptr = idpoint;
				break;
			}
		}
		return ptr;
	}

	void ambushOn(Actor mo, Actor ptr)
	{
		bool am_shadows = CVar.FindCvar("am_shadows").GetBool();
		bool am_debug = CVar.FindCvar("am_debug").GetBool();

		if (mo.args[AMSTATUS] == 0)
		{
			mo.args[AMSTATUS] = 1;
			mo.target = ptr;
			mo.bNOTARGETSWITCH = true;
			mo.reactiontime = -1;
			if (am_shadows)
			{
				if (am_debug)
				{
					mo.SetShade("Blue");
					mo.A_SetRenderStyle(0.5, STYLE_Stencil);
				}
				else
				{
					mo.A_SetRenderStyle(0.5, STYLE_Subtract);
				}
			}
		}
	}
	void ambushOff(Actor mo)
	{
		bool am_shadows = CVar.FindCvar("am_shadows").GetBool();

		if (mo.args[AMSTATUS] == 1) // dormant
		{
			mo.args[AMSTATUS] = 0;
			mo.target = null;
			mo.bNOTARGETSWITCH = false;
			mo.SeeSound = "";
			mo.reactiontime = 8;
			if (am_shadows)
			{
				mo.A_SetRenderStyle(1.0, STYLE_Normal);
			}
		}
		if (mo.args[AMSTATUS] == 2) // targeting player
		{
			mo.args[AMSTATUS] = 0;
			mo.target = player.mo;
			mo.bNOTARGETSWITCH = false;
			mo.reactiontime = 8;
			if (am_shadows)
			{
				mo.A_SetRenderStyle(1.0, STYLE_Normal);
			}
		}
	}

	void moambush()
	{
		ThinkerIterator MonsterFinder;
		MonsterFinder = ThinkerIterator.Create("Actor", Thinker.STAT_DEFAULT);
		Actor mo;

		int am_radius = CVar.FindCvar("am_radius").GetInt();

		while (mo = Actor(MonsterFinder.Next()))
		{
			if (mo.bIsMonster && mo.args[MOFLAG] == 2)
			{
				if (mo.Distance2D(player.mo) < am_radius)
				{
					if (mo.CurState == mo.ResolveState("See"))
					{
						AM_Point ptr = findambush(mo, player.mo);
						If (ptr)
						{
							ambushOn(mo, ptr);
						}
					}
					else if (mo.args[AMSTATUS] == 1 && mo.CheckSight(player.mo) && mo.Distance2D(player.mo) < 256)
					{
						mo.args[AMSTATUS] = 2;
						ambushOff(mo);
					}
					else if (!mo.CheckSight(player.mo))
					{
						ambushOff(mo);
					}
				}
				else
				{
					ambushOff(mo);
				}
			}
		}
	}

	override void WorldTick()
	{
		bool am_enabled = CVar.FindCvar("am_enabled").GetBool();
		if (am_enabled)
		{
			moambush();
		}
	}

	// return AM_Point spawned
	Actor spawnPoint(vector3 v3, Actor center)
	{
		Actor ptr = Actor.Spawn("AM_Point", v3);
		// remove if it can be seen from the center (will eliminate most convex corners)
		if (center.CheckSight(ptr))
		{
			ptr.A_Remove(AAPTR_DEFAULT, RMVF_EVERYTHING, "AM_Point");
			ptr = null;
		}
		else
		{
			bool am_debug = CVar.FindCvar("am_debug").GetBool();
			if (am_debug)
			{
				ptr.scale = (1.0, 1.0);
			}
		}

		return ptr;
	}

	bool sharpCorner(double vabs)
	{
		int am_angle = CVar.FindCvar("am_angle").GetInt();
		return ((vabs > 90 - am_angle && vabs < 90 + am_angle) || (vabs > 270 - am_angle && vabs < 270 + am_angle));
	}

	// add sector corner points
	override void WorldLoaded(WorldEvent e)
	{
		int BLOCKING = 1;                 // bit 0x0001
		int TWOSIDED = 2;                 // bit 0x0004
		int DONTPEGTOP = 4;               // bit 0x0008
		int DONTPEGBOTTOM = 8;            // bit 0x0010
		int INNER = 0;
		
		MOFLAG = 0;
		AMSTATUS = 1;

		for(int i = 0; i < Level.Sectors.Size(); i++)
		{
			SecPlane floorplane = Level.Sectors[i].FloorPlane;
			int floorZ = int(floorplane.ZAtPoint((Level.Sectors[i].CenterSpot.x, Level.Sectors[i].CenterSpot.y)));

			vector3 v3 = (Level.Sectors[i].CenterSpot.x, Level.Sectors[i].CenterSpot.y, floorZ);
			
			Actor center = Actor.Spawn("AM_Point", v3);

			for (int ii = 0; ii < Level.Sectors[i].Lines.Size(); ii++)
			{
				vector2 vA1 = Level.Sectors[i].Lines[ii].V1.P;
				vector2 vA2 = Level.Sectors[i].Lines[ii].V2.P;
				double vaii = VectorAngle(vA1.x - vA2.x, vA1.y - vA2.y);
				int AB = 0;
				
				int f = Level.Sectors[i].lines[ii].flags;
				int peg_top = f & (TWOSIDED + DONTPEGTOP);
				int peg_bottom = f & (TWOSIDED + DONTPEGBOTTOM);
				int blocking = f & (BLOCKING);

				Side sidedef = Level.Sectors[i].Lines[ii].Sidedef[INNER];
				textureid botid = sidedef.GetTexture(sidedef.Bottom); // depends on good design
				bool hide = !botid.isNull();

				if (blocking || hide)
				{
					for (int j = 0; j < Level.Sectors[i].Lines.Size(); j++)
					{
						if (j != ii)
						{
							vector2 vB1 = Level.Sectors[i].Lines[j].V1.P;
							vector2 vB2 = Level.Sectors[i].Lines[j].V2.P;
							double vaj = VectorAngle(vB1.x - vB2.x, vB1.y - vB2.y);
							double vabs = abs(abs(vaii) - abs(vaj));

							if (vA1 == vB2 && sharpCorner(vabs))
							{
								v3 = (vA1.x, vA1.y, floorZ);
								spawnPoint(v3, center);
								AB++;
							}
							if (vA2 == vB1 && sharpCorner(vabs))
							{
								v3 = (vA2.x, vA2.y, floorZ);
								spawnPoint(v3, center);
								AB++;
							}
						}
						if (AB == 2)
						{
							break;
						}
					}
				}
			}
			center.A_Remove(AAPTR_DEFAULT, RMVF_EVERYTHING, "AM_Point");
			center = null;
		}
	}

	override void WorldThingSpawned(WorldEvent e)
	{
		if (e.thing && e.thing.bIsMonster && e.thing.args[MOFLAG] != 1)
		{
			int am_flankerchance = CVar.FindCvar("am_flankerchance").GetInt();
			e.thing.args[MOFLAG] = Random(2, am_flankerchance);
			e.thing.args[AMSTATUS] = 0;
		}
	}
	override void PlayerRespawned(PlayerEvent e)
	{
		player = players[e.PlayerNumber];
	}
	override void PlayerSpawned(PlayerEvent e)
	{
		player = players[e.PlayerNumber];
	}
}

class ambushers_OptionMenu : OptionMenu
{
	override void Init(Menu parent, OptionMenuDescriptor desc)
	{
		super.Init(parent, desc);
		DontDim = true;
		DontBlur = true;
	}
}
